/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.jaccal.util.NumUtil;

import com.simfill.LogManager;
import com.simfill.Registry;

import java.lang.StringBuffer;
import java.lang.String;

// Stores the data to be stored in a record of an extension file (EF_EXT1, EF_EXT2, EF_EXT3)
public class Extension extends LinearFixedFile
{
	public static final byte EXT_TYPE_UNKNOWN = 0x00;
	public static final byte EXT_TYPE_SUBADR  = 0x01;
	public static final byte EXT_TYPE_ADDDATA = 0x02;
	
	public Extension( short size, byte type )
	{
		super( size );
		setType( type );
		setData( "" );
		setExtension( ( byte )0xff );
	}
	
	public void setType( byte type )
	{
		m_type = type;
	}

	public void setData( String data )
	{
		m_data = data;
	}
	
	public void setExtension( byte extension )
	{
		m_extension = extension;
	}
	
	public String toHex()
	{
		byte [] storage = new byte [ getSize() ];
		int idx = 0;
		
		storage[ idx++ ] = m_type;
		
		switch( m_type )
		{
			case EXT_TYPE_ADDDATA:
					
				byte maxNumSize = ( byte )((  getSize() - 3 ) << 1 );
				byte numSize = m_data.length() < ( int )maxNumSize ? ( byte )m_data.length() : maxNumSize;

				if( numSize < m_data.length() )
				{
					LogManager.getInstance().log( "warn", "The maximum size of an extension number is " + maxNumSize + ". The entry '" + m_data + "' will be truncated. You might want to consider using an other extension record for longer number.", true );
				}
				
				storage[ idx++ ] = ( byte )( ( numSize >> 1 ) + ( numSize & 1 ) );
				for( byte digit = 0; digit < numSize; digit += 2 )
				{
					byte dl = ( byte )( m_data.charAt( digit ) - '0' );
					byte dh = ( byte )( digit + 1 < numSize ? ( byte )m_data.charAt( digit + 1 ) - '0' : 0xff );
					storage[ idx++ ] = ( byte )( dl + ( dh << 4 ) );
				}
				while( idx < getSize() - 1 )
				{
					storage[ idx++ ] = ( byte )0xff;
				}
	
				break;
				
			default:
					
				LogManager.getInstance().log( "err", "Unsupported record type. Blank data will be written.", true );
			
				while( idx < getSize() - 1 )
				{
					storage[ idx++ ] = ( byte )0xff;
				}
		}

		storage[ idx++ ] = m_extension;
		
		return NumUtil.hex2String( storage );
	}
	
	public String toString()
	{
		StringBuffer report = new StringBuffer();
		
		String newline = Registry.getInstance().get( "report.newline" );

		report.append( "Type of record: " );
		switch( m_type )
		{
			case EXT_TYPE_UNKNOWN:		report.append( "unknown (unsupported)"					); break;
			case EXT_TYPE_SUBADR:		report.append( "called party subaddress (unsupported)"	); break;
			case EXT_TYPE_ADDDATA:		report.append( "additional data"						); break;
			default:					report.append( "unsupported"							);
		}
		report.append( newline );
		
		if( m_type == EXT_TYPE_ADDDATA )
		{
			report.append( "Data: " );
			report.append( m_data.length() <= 20 ? m_data : m_data.substring( 0, 20 ) );
			report.append( newline );
		}
		
		report.append( "Extension: " );
		report.append( m_extension == ( byte )0xff ? "none" : m_extension );
		report.append( newline );
		
		return report.toString();
	}

	// Record type
	private byte m_type;
	
	// Extension data
	private String m_data;
	
	// Extension record identifier - see EF_EXT?
	private byte m_extension;
}
