/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill;

import com.jaccal.CardException;
import com.jaccal.CardResponse;
import com.jaccal.Session;
import com.jaccal.SessionFactory;
import com.jaccal.command.Command;
import com.jaccal.command.gsm.GSMGetResponse;
import com.jaccal.command.gsm.GSMSelect;
import com.jaccal.command.gsm.GSMUpdateBinary;
import com.jaccal.command.gsm.GSMUpdateRecord;
import com.jaccal.command.gsm.GSMVerifyChvEx;
import com.jaccal.command.gsm.axalto.AxaltoVerifyKey;
import com.jaccal.util.NumUtil;
import com.simfill.gsm.FileInfo;
import com.simfill.gsm.LinearFixedFile;
import com.simfill.gsm.RawData;
import com.simfill.gsm.RawRecord;
import com.simfill.gsm.TransparentFile;

// Wrapper class for the Jaccal API
public class PCSC
{
	public static final String MF          = "3F00";
	public static final String DF_TELECOM  = "7F10";
	public static final String DF_GSM      = "7F20";
	public static final String DF_DCS1800  = "7F21";
	public static final String EF_ELP      = "2F05";
	public static final String EF_ICCID    = "2FE2";
	public static final String EF_LP       = "6F05";
	public static final String EF_IMSI     = "6F07";
	public static final String EF_PLMNsel  = "6F30";
	public static final String EF_SST      = "6F38";
	public static final String EF_ADN      = "6F3A";
	public static final String EF_FDN      = "6F3B";
	public static final String EF_SMS      = "6F3C";
	public static final String EF_MSISDN   = "6F40";
	public static final String EF_LND      = "6F44";
	public static final String EF_SPN      = "6F46";
	public static final String EF_SDN      = "6F49";
	public static final String EF_EXT1     = "6F4A";
	public static final String EF_EXT2     = "6F4B";
	public static final String EF_EXT3     = "6F4C";
	public static final String EF_EXT4     = "6F4E";
	public static final String EF_LOCIGPRS = "6F53";
	public static final String EF_FPLMN    = "6F7B";
	public static final String EF_LOCI     = "6F7E";
	public static final String EF_AD       = "6FAD";
	public static final String EF_PHASE    = "6FAE";
	
	public static final byte STATUS_SUCCESS   = ( byte )0x90;
	public static final byte STATUS_MOREDATA  = ( byte )0x9F;
	public static final byte STATUS_ERROR_MEM = ( byte )0x92;
	public static final byte STATUS_ERROR_REF = ( byte )0x94;
	public static final byte STATUS_ERROR_SEC = ( byte )0x98;

	public static final byte STATUS_ERROR_SEC_NOTINIT   = ( byte )0x02;
	public static final byte STATUS_ERROR_SEC_FAILED    = ( byte )0x04;
	public static final byte STATUS_ERROR_SEC_CHVSTATUS = ( byte )0x08;
	public static final byte STATUS_ERROR_SEC_INVSTATUS = ( byte )0x10;
	public static final byte STATUS_ERROR_SEC_LOCKED    = ( byte )0x40;
	public static final byte STATUS_ERROR_SEC_MAXREACH  = ( byte )0x50;
	
	public enum CHV { PIN1, PIN2, ADM1 }

	public PCSC() throws Exception
	{
		m_session = SessionFactory.getInstance().createSession( "default" );
	}
	
	public FileInfo getFileInfo()
	{
		return m_finfo;
	}
	
	public void shutdown() throws Exception
	{
		m_session.close();
	} 
	
	// Authenticate to the smartcard using CHV (PIN1, PIN2, ADM1, etc.)
	public void authenticate( CHV chv, String code ) throws Exception
	{
		String chvIndex = null;
		String chvLabel = null;
		
		switch( chv )
		{
			case PIN1:	chvIndex = GSMVerifyChvEx.CHV1; chvLabel = "PIN1"; break;
			case PIN2:	chvIndex = GSMVerifyChvEx.CHV2; chvLabel = "PIN2"; break;
			case ADM1:	chvIndex = GSMVerifyChvEx.ADM1; chvLabel = "ADM1"; break;
		}

		CardResponse r = execute( new GSMVerifyChvEx( chvIndex, code ) );
		switch( r.statusWord.getSw1() )
		{
			case STATUS_SUCCESS:
				
				LogManager.getInstance().log( "info", "Authentication: " + chvLabel + " accepted.", true );
				break;
				
			case STATUS_ERROR_SEC:
				
				switch( r.statusWord.getSw2() )
				{
					case STATUS_ERROR_SEC_NOTINIT:
						LogManager.getInstance().log( "warn", "Authentication: " + chvLabel + " not initialized.", true );
						break;
					case STATUS_ERROR_SEC_FAILED:
						LogManager.getInstance().log( "err", "Authentication: " + chvLabel + " is not correct. At least one attempt left.", true );
						break;
					case STATUS_ERROR_SEC_CHVSTATUS:
						LogManager.getInstance().log( "warn", "Authentication: " + chvLabel + " is not activated.", true );
						break;
					case STATUS_ERROR_SEC_INVSTATUS:
						LogManager.getInstance().log( "err", "Authentication: the selected file is invalidated.", true );
						break;
					case STATUS_ERROR_SEC_LOCKED:
						LogManager.getInstance().log( "err", "Authentication: " + chvLabel + " is not correct. No attempt left.", true );
						break;
					case STATUS_ERROR_SEC_MAXREACH:
						LogManager.getInstance().log( "err", "Authentication: increase can not be performed. Maximum value reached.", true );
						break;
					default:
						LogManager.getInstance().log( "err", "Authentication: SIM generated an unhandled security error while sending " + chvLabel + ".", true );
				}
				break;
				
			default:
					
				LogManager.getInstance().log( "err", "Authentication: SIM generated an unhandled error while sending " + chvLabel + ".", true );
		}
	}
	
	// Authenticate to an Axalto smartcard using an ADM key
	public void authenticateAxalto( String keyRefId, String key ) throws Exception
	{
		CardResponse r = execute( new AxaltoVerifyKey( ( byte )0, Byte.parseByte( keyRefId ), key ) );
		
		switch( r.statusWord.getSw1() )
		{
			case STATUS_SUCCESS:
				
				LogManager.getInstance().log( "info", "Authentication: ADM code accepted.", true );
				break;
				
			case STATUS_ERROR_SEC:
				
				switch( r.statusWord.getSw2() )
				{
					case STATUS_ERROR_SEC_NOTINIT:
						LogManager.getInstance().log( "warn", "Axalto authentication: ADM code not initialized.", true );
						break;
					case STATUS_ERROR_SEC_FAILED:
						LogManager.getInstance().log( "err", "Axalto authentication: ADM code is not correct. At least one attempt left.", true );
						break;
					case STATUS_ERROR_SEC_CHVSTATUS:
						LogManager.getInstance().log( "warn", "Axalto authentication: ADM code is not activated.", true );
						break;
					case STATUS_ERROR_SEC_INVSTATUS:
						LogManager.getInstance().log( "err", "Axalto authentication: the selected file is invalidated.", true );
						break;
					case STATUS_ERROR_SEC_LOCKED:
						LogManager.getInstance().log( "err", "Axalto authentication: ADM code is not correct. No attempt left.", true );
						break;
					case STATUS_ERROR_SEC_MAXREACH:
						LogManager.getInstance().log( "err", "Axalto authentication: increase can not be performed. Maximum value reached.", true );
						break;
					default:
						LogManager.getInstance().log( "err", "Axalto authentication: SIM generated an unhandled security error while sending ADM code.", true );
				}
				break;
				
			default:
					
				LogManager.getInstance().log( "err", "Axalto authentication: SIM generated an unhandled error while sending ADM code.", true );
		}
	}

	public void select( String fid ) throws CardException
	{
		try
		{
			m_finfo = null;
	
			// Send the select command
			CardResponse r = execute( new GSMSelect( fid ) );
	
			// Check the response status
			if( r.statusWord.getSw1() == STATUS_MOREDATA )
			{
				// Get the result of the select command
				r = execute( new GSMGetResponse( r ) );
					
				// Check the response status
				if( r.statusWord.getSw1() == STATUS_SUCCESS )
				{
					m_finfo = new FileInfo( r.getData() );
				}
				else
				{
					throw new CardException( "select( '" + fid + "' ): status '" + r.statusWord + "' (expected status 'SUCCESS')." );
				}
			}
			else
			{
				throw new CardException( "select( '" + fid + "' ): status '" + r.statusWord + "' (expected status 'MORE DATA')." );
			}
		}
		catch( CardException e )
		{
			throw new CardException( "Error selecting the file '" + fid + "': " + e.toString() );
		}
	}
	
	public void select( String [] path ) throws Exception
	{
		for( String fid : path )
		{
			select( fid );
		}
	}

	public void update( TransparentFile data ) throws Exception
	{
		updateBinary( data, true );
	}
	
	public void update( int recNum, LinearFixedFile data ) throws Exception
	{
		updateRecord( recNum, data, true );
	}

	public void reset() throws CardException
	{
		if( m_finfo == null )
		{
			throw new CardException( "resetFile: no file previously selected." );
		}

		if( m_finfo.getType() != FileInfo.FILETYPE_EF )
		{
			throw new CardException( "resetFile( '" + m_finfo.getSID() + "' ): the file is not an 'elementary file'." );
		}

		switch( m_finfo.getStruct() )
		{
			case FileInfo.STRUCT_CYCLIC:
			case FileInfo.STRUCT_LINEARFIXED:
			{
				byte [] data = new byte[ m_finfo.getRecSize() ];
				for( int it = 0; it < m_finfo.getRecSize(); ++it )
				{
					data[ it ] = ( byte )0xff;
				}
				reset( NumUtil.hex2String( data ) );
				break;
			}
				
			case FileInfo.STRUCT_TRANSPARENT:
			{
				byte [] data = new byte[ m_finfo.getSize() ];
				for( int it = 0; it < m_finfo.getSize(); ++it )
				{
					data[ it ] = ( byte )0xff;
				}
				reset( NumUtil.hex2String( data ) );
				break;
			}
		}
	}
	
	public void reset( String data ) throws CardException
	{
		if( m_finfo == null )
		{
			throw new CardException( "resetFile: no file previously selected." );
		}

		if( m_finfo.getType() != FileInfo.FILETYPE_EF )
		{
			throw new CardException( "resetFile( '" + m_finfo.getSID() + "' ): the file is not an 'elementary file'." );
		}

		switch( m_finfo.getStruct() )
		{
			case FileInfo.STRUCT_CYCLIC:
			case FileInfo.STRUCT_LINEARFIXED:

				RawRecord rec = new RawRecord( m_finfo.getRecSize(), data );
				for( int it = 1; it <= m_finfo.getSize() / m_finfo.getRecSize(); ++it )
				{
					updateRecord( it, rec, false );
				}
				break;
				
			case FileInfo.STRUCT_TRANSPARENT:

				updateBinary( new RawData( m_finfo.getSize(), data ), false );
				break;
		}
	}

	private void updateBinary( TransparentFile data, boolean doLog ) throws CardException
	{
		if( m_finfo == null )
		{
			throw new CardException( "resetFile: no file previously selected." );
		}

		// Check the file structure
		if( m_finfo.getStruct() != FileInfo.STRUCT_TRANSPARENT )
		{
			throw new CardException( "updateBinary( '" + m_finfo.getSID() + "' ): the structure of the file is not 'transparent'." );
		}	

		// Check the size of the data
		String hex = data.toHex();
		if( hex.length() != m_finfo.getSize() << 1 )
		{
			throw new CardException( "updateBinary( '" + m_finfo.getSID() + "' ): the length of the provided data does not match the size of the file." );
		}
		
		// Update the binary file
		CardResponse r = execute( new GSMUpdateBinary( "00", hex ) );

		if( r.statusWord.getSw1() != STATUS_SUCCESS )
		{
			LogManager.getInstance().log( "err", "<li>", false );
			LogManager.getInstance().log( "err", "The following data has not been written:", true );
			LogManager.getInstance().log( "err", data, true );
			LogManager.getInstance().log( "err", "</li>", false );

			throw new CardException( "updateBinary( '" + m_finfo.getSID() + "' ): status '" + r.statusWord + "' (expected status 'SUCCESS')." );
		}

		// Report written data
		if( doLog )
		{
			LogManager.getInstance().log( "rprt", "<li>" + data + "</li>", true );
		}
	}

	private void updateRecord( int recNum, LinearFixedFile data, boolean doLog ) throws CardException
	{
		if( m_finfo == null )
		{
			throw new CardException( "resetFile: no file previously selected." );
		}

		// Check the file structure
		if( m_finfo.getStruct() != FileInfo.STRUCT_LINEARFIXED && m_finfo.getStruct() != FileInfo.STRUCT_CYCLIC )
		{
			throw new CardException( "updateRecord( '" + m_finfo.getSID() + "' ): the structure of the file is not 'linear fixed' or 'cyclic'." );
		}	

		// Check the size of the data
		String hex = data.toHex();
		if( hex.length() != m_finfo.getRecSize() << 1 )
		{
			throw new CardException( "updateRecord( '" + m_finfo.getSID() + "' ): the length of the provided data does not match the size of a record." );
		}
		
		String index;
		byte mode;

		if( m_finfo.getStruct() == FileInfo.STRUCT_CYCLIC )
		{
				index = "00";
				mode = GSMUpdateRecord.PREV_REC;
		}
		else
		{
				index = NumUtil.int2HexString( recNum ).substring( 6, 8 );
				mode = GSMUpdateRecord.CURRENT_REC;		
		}

		// Update the record
		CardResponse r = execute( new GSMUpdateRecord( index, hex, mode ) );
			
		if( r.statusWord.getSw1() != STATUS_SUCCESS )
		{
			LogManager.getInstance().log( "err", "<li>", false );
			LogManager.getInstance().log( "err", "The following data has not been written:", true );
			LogManager.getInstance().log( "err", data, true );
			LogManager.getInstance().log( "err", "</li>", false );

			throw new CardException( "updateRecord( '" + m_finfo.getSID() + "' ): status '" + r.statusWord + "' (expected status 'SUCCESS')." );
		}

		// Report written data
		if( doLog )
		{
			LogManager.getInstance().log( "rprt", "<li>" + data + "</li>", true );
		}
	}
	
	// Execute a command and log the request and result
	private CardResponse execute( Command cmd ) throws CardException
	{
		LogManager.getInstance().log( "adpu", cmd, true );
		
		CardResponse r = m_session.execute( cmd );
		
		LogManager.getInstance().log( "adpu", r, true );
		
		return r;
	}
	
	// PC/SC session
	private Session m_session;
	
	// Information on the currently selected file
	private FileInfo m_finfo;
}