/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.jaccal.util.NumUtil;

import com.simfill.LogManager;
import com.simfill.Registry;

import java.lang.StringBuffer;
import java.lang.String;
import java.util.StringTokenizer;

public class SMSDeliver extends SMSUserData implements SMSReplyPath
{
	// TP-More-Messages-to-Send (TP-MMS), according to 3GPP TS 23.040
	// Bit 2 of the first byte of a PDU.
	public static final byte MMS_MORE_MESSAGES    = 0x00;
	public static final byte MMS_NO_MORE_MESSAGES = 0x04;
	
	protected byte m_mms;
	
	// TP-Reply-Path (TP-RP), according to 3GPP TS 23.040
	// Bit 7 of the first byte of a PDU.
	protected byte m_rp;
	
	// TP-Status-Report-Indicator (TP-SRI), according to 3GPP TS 23.040
	// Bit 5 of the first byte of a PDU.
	public static final byte SRI_DONT_REPORT      = 0x00;
	public static final byte SRI_DO_REPORT        = 0x20;
	
	protected byte m_sri;
	
	// TP-Originating-Address (TP-OA), according to 3GPP TS 23.040
	protected Address m_oa;
	
	// TP-Service-Center-Time-Stamp (TP-SCTS), according to 3GPP TS 23.040
	protected byte [] m_scts;
	protected String m_timedate;
	
	public SMSDeliver( short size ) throws Exception
	{
		super( size );
		
		setMTI( MTI_SC2MS_SMS_DELIVER );
		setMMS( MMS_NO_MORE_MESSAGES );
		setRP ( RP_NOT_SET );
		setSRI( SRI_DONT_REPORT );
		setOA ( new Address() );
		m_scts = new byte[ 7 ];
		m_timedate = "";
	}
	
	public void setMMS( byte mms )
	{
		m_mms = mms;
	}
	
	public void setRP( byte rp )
	{
		m_rp = rp;
	}
	
	public void setSRI( byte sri )
	{
		m_sri = sri;
	}
	
	public void setOA( Address oa )
	{
		m_oa = oa;
	}

	public void setSCTS( String timedate )
	{
		m_timedate = timedate;

		// Parse the date manually -_-
		// TODO: validate the format of the string
		StringTokenizer st = new StringTokenizer( timedate );
		int year     = Integer.valueOf( st.nextToken( "-"  ) ) % 100;
		int month    = Integer.valueOf( st.nextToken( "-"  ) );
		int day      = Integer.valueOf( st.nextToken( "- " ) );
		int hour     = Integer.valueOf( st.nextToken( ": " ) );
		int minute   = Integer.valueOf( st.nextToken( ":"  ) );
		int second   = Integer.valueOf( st.nextToken( ": " ) );
		int timezone = Integer.valueOf( st.nextToken( " +" ) );
		timezone = ( 60 * ( ( timezone / 100 ) % 100 ) ) + ( timezone % 100 );
		
		timezone /= 15;
		timezone = ( ( ( timezone / 10 ) % 10 ) << 4 ) + ( timezone % 10 );
		timezone = timezone < 0 ? -timezone | 0x80 : timezone;

		m_scts[ 0 ] = ( byte )( ( (     year % 10   ) << 4 ) + ( (     year / 10   ) % 10 ) );
		m_scts[ 1 ] = ( byte )( ( (    month % 10   ) << 4 ) + ( (    month / 10   ) % 10 ) );
		m_scts[ 2 ] = ( byte )( ( (      day % 10   ) << 4 ) + ( (      day / 10   ) % 10 ) );
		m_scts[ 3 ] = ( byte )( ( (     hour % 10   ) << 4 ) + ( (     hour / 10   ) % 10 ) );
		m_scts[ 4 ] = ( byte )( ( (   minute % 10   ) << 4 ) + ( (   minute / 10   ) % 10 ) );
		m_scts[ 5 ] = ( byte )( ( (   second % 10   ) << 4 ) + ( (   second / 10   ) % 10 ) );
		m_scts[ 6 ] = ( byte )( ( ( timezone & 0xf0 ) >> 4 ) + ( ( timezone & 0x0f ) << 4 ) );
	}

	public String toHex()
	{
		byte [] storage = new byte[ getSize() ];

		int idx = 0;

		// Store the status of SMS
		storage[ idx++ ] = m_status;

		// Store the service center number if provided
		String number = m_sca.getNumber();
		if( number.length() == 0 )
		{
			storage[ idx++ ] = 0;
		}
		else
		{
			int size = number.length();
			
			// Check the size of the number
			if( size > 20 )
			{
				LogManager.getInstance().log( "warn", "SMSDeliver: The length of the service center address is limited to 20 digits. The number '" + number + "' will be truncated.", true );
				size = 20;
			}
			
			storage[ idx++ ] = ( byte )( ( size >> 1 ) + ( size & 1 ) + 1 );
			storage[ idx++ ] = m_sca.getTypeOfAddress();
			number += ( char )0xff;
			for( int ch = 0; ch < size; ch += 2 )
			{
				storage[ idx++ ] = ( byte )( ( number.charAt( ch ) - '0' ) + ( ( number.charAt( ch + 1 ) - '0' ) << 4 ) );
			}
		}

		// Store the flags in the first byte of the TDPU
		storage[ idx++ ] = ( byte )( m_mti | m_mms | m_sri | m_udhi | m_rp );

		// Store the originated address if provided
		number = m_oa.getNumber();
		if( number.length() == 0 )
		{
			storage[ idx++ ] = 0;
			storage[ idx++ ] = m_oa.getTypeOfAddress();
		}
		else
		{
			int size = number.length();
			
			// Check the size of the number
			if( size > 20 )
			{
				LogManager.getInstance().log( "warn", "SMSDeliver: The length of the originated address is limited to 20 digits. The number '" + number + "' will be truncated.", true );
				size = 20;
			}
			
			storage[ idx++ ] = ( byte )size;
			storage[ idx++ ] = m_oa.getTypeOfAddress();
			number += ( char )0xff;
			for( int ch = 0; ch < size; ch += 2 )
			{
				storage[ idx++ ] = ( byte )( ( number.charAt( ch ) - '0' ) + ( ( number.charAt( ch + 1 ) - '0' ) << 4 ) );
			}
		}

		// Store the protocole identifier
		storage[ idx++ ] = m_pid;

		// Store the data coding scheme
		storage[ idx++ ] = m_dcs;

		// Store the service center timestamp
		for( byte b : m_scts )
		{
			storage[ idx++ ] = b;
		}

		// Store the user data header, if necessary
		int udl_idx = idx++;
		
		if( m_udhi == UDHI_HEADER_PRESENT )
		{
			storage[ idx++ ] = ( byte )m_udhl;

			for( int it = 0; it < m_udh.size() && idx < getSize(); ++it )
			{
				SMSUserDataHeaderElement e = m_udh.elementAt( it );
				
				System.arraycopy( e.getData(), 0, storage, idx, e.getSize() );
				idx += e.getSize();
			}
		}
		
		// Store the user data
		try
		{
			int avail;

			switch( m_dcs )
			{
				case DCS_16BITS_UCS2:
					
					// Check the space available
					avail = Math.min( 140 - ( m_udhl + 1 ), getSize() - idx ) >> 1;
					
					// Check the length of the text
					if( m_udl > avail )
					{
						m_udl = avail;
						LogManager.getInstance().log( "warn", "SMSDeliver: The text '" + m_ud + "' is UCS2-encoded. It will be truncated to " + avail + " characters.", true );
					}

					// Store the text as is, as Java stores the text UCS2 encoded already
					for( int ch = 0; ch < m_udl; ++ch )
					{
						char c = m_ud.charAt( ch );

						storage[ idx++ ] = ( byte )( ( c & 0xff00 ) >> 8 );
						storage[ idx++ ] = ( byte )( ( c & 0x00ff ) >> 0 );
					}
					
					// Store the size of the user data, including the header
					storage[ udl_idx ] = ( byte )( idx - udl_idx - 1 );

					break;

				case DCS_7BITS_DEFAULT:

					int offset = 0;
					int boundary = 0;

					// Compute calculation related to 7-bit boundary padding.
					// According to 3GPP TS 23.040, the text has to start on a 7-bit boundary. 
					if( m_udhi == UDHI_HEADER_PRESENT )
					{
						offset = ( m_udhl + 1 ) << 3;								// number of bits used by the headers
						boundary = 7 * ( int )Math.ceil( ( double )offset / 7.0 );	// number of bits to reach the next 7-bit boundary
						avail = Math.min( 160 - boundary / 7, getSize() - idx );	// "avail" is the number of septets available
						offset -= boundary - 7;
					}
					else
					{
						avail = Math.min( 160, getSize() - idx );
					}

					if( m_udl > 0 )
					{
						// Check the length of the text
						byte [] trText = Charset.translate( m_ud );
						m_udl = trText.length;
						if( m_udl > avail )
						{
							m_udl = avail;
							LogManager.getInstance().log( "warn", "SMSDeliver: The text '" + m_ud + "' is 7-bits-encoded and will be truncated.", true );
						}
	
						// If the truncation happens on a 2-bytes special character, also remove the remaining byte
						if( trText[ m_udl - 1 ] == Charset.CHARSET_ESCAPE ) --m_udl;
	
						// Encode the text using the 8 bits to 7 bits compression algorithm
						byte [] padding = { 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f };
						byte [] padText = new byte[ m_udl + padding.length ];
						System.arraycopy( trText,  0, padText, 0, m_udl );
						System.arraycopy( padding, 0, padText, m_udl, padding.length );
	
						int ch = 0;
						int ch1 = ( m_udhi == UDHI_HEADER_PRESENT ) ? 0x7f : padText[ ch++ ]; // Pad the unused bits between the end of the header and the next 7-bit boundary where the text begins
						while( ( ch <= m_udl || offset > 0 ) && idx < getSize() )
						{
							int ch2 = padText[ ch++ ];
							storage[ idx++ ] = ( byte )( ( ch1 >> offset ) + ( ( ch2 & ( 0x000000ff >> ( 7 - offset ) ) ) << ( 7 - offset ) ) );
							if( ++offset < 7 )
							{
								ch1 = ch2;
							}
							else
							{
								offset = 0;
								ch1 = padText[ ch++ ];
							}
						}

						// Store the number of septets of the user data, including the header
						storage[ udl_idx ] = ( byte )Math.ceil( ( boundary + m_udl * 7.0 ) / 7 );
						//storage[ udl_idx ] = ( byte )Math.floor( ( ( idx - udl_idx - 1 ) * 8 ) / 7 );

						// Store the truncated user data
						setUD( Charset.untranslate( trText, 0, m_udl ) );
					}
					else
					{
						// Store the number of septets of the user data, including the header
						storage[ udl_idx ] = ( byte )( boundary / 7 );
					}

					break;
					
				default:
					
					LogManager.getInstance().log( "warn", "SMSDeliver: unsupported data coding scheme (TP-DCS). No user data will be stored in the SMS message.", true );

					// Store the size of the user data, including the header
					storage[ udl_idx ] = ( byte )( idx - udl_idx - 1 );

					break;
			}
		}
		catch( Exception e )
		{
			LogManager.getInstance().log( "excp", e, true );
			return null;
		}

		// Fill the remaining space with padding
		while( idx < getSize() )
		{
			storage[ idx++ ] = -1;
		}

		return NumUtil.hex2String( storage );
	}
	
	public String toString()
	{
		StringBuffer report = new StringBuffer();
		
		String newline = Registry.getInstance().get( "report.newline" );
		
		report.append( "Status: " );
		switch( m_status )
		{
			case STATUS_FREESPACE:    report.append( "Empty / deleted"    ); break;
			case STATUS_IN_READ:      report.append( "Incoming, read"     ); break;
			case STATUS_IN_NOT_READ:  report.append( "Incoming, not read" ); break;
			case STATUS_OUT_SENT:     report.append( "Outgoing, sent"     ); break;
			case STATUS_OUT_NOT_SENT: report.append( "Outgoing, not sent" ); break;
			default:                  report.append( "INVALID" );
		}
		report.append( newline );
		
		report.append( "Service center address: " );
		report.append( newline );
		if( m_sca.getNumber().length() > 20 )
		{
			Address a = new Address();
			a.setTON( m_sca.getTON() );
			a.setNPI( m_sca.getNPI() );
			a.setNumber( m_sca.getNumber().substring( 0, 20 ) );
			report.append( a );
		}
		else
		{
			report.append( m_sca );
		}
		
		report.append( "Type: " );
		switch( m_mti )
		{
			case MTI_SC2MS_SMS_DELIVER:        report.append( "SMS-DELIVER"       ); break;
			case MTI_SC2MS_SMS_SUBMIT_REPORT:  report.append( "SMS-SUBMIT-REPORT" ); break;
			case MTI_SC2MS_SMS_STATUS_REPORT:  report.append( "SMS-STATUS-REPORT" ); break;
			case MTI_RESERVED:                 report.append( "Reserved"          ); break;
			default:                           report.append( "INVALID"           );
		}
		report.append( newline );
		
		report.append( "More messages: " );
		report.append( m_mms == MMS_MORE_MESSAGES );
		report.append( newline );
		
		report.append( "Status report requested: " );
		report.append( m_sri == SRI_DO_REPORT );
		report.append( newline );
		
		report.append( "User data header present: " );
		report.append( m_udhi == UDHI_HEADER_PRESENT );
		report.append( newline );
		
		report.append( "Reply path set: " );
		report.append( m_rp == RP_SET );
		report.append( newline );
		
		report.append( "Originated address: ");
		report.append( newline );
		if( m_oa.getNumber().length() > 20 )
		{
			Address a = new Address();
			a.setTON( m_oa.getTON() );
			a.setNPI( m_oa.getNPI() );
			a.setNumber( m_oa.getNumber().substring( 0, 20 ) );
			report.append( a );
		}
		else
		{
			report.append( m_oa );
		}
		
		report.append( "Protocol identifier: NOT SUPPORTED" ); // FIXME
		report.append( newline );
		
		report.append( "Data coding scheme: " );
		switch( m_dcs )
		{
			case DCS_7BITS_DEFAULT: report.append( "7 bits (default)" ); break;
			case DCS_8BITS:         report.append( "8 bits"           ); break;
			case DCS_16BITS_UCS2:   report.append( "16 bits (UCS2)"   ); break;
			case DCS_RESERVED:      report.append( "Reserved"         ); break;
			default:                report.append( "INVALID"          );
		}
		report.append( newline );
		
		report.append( "Service center timestamp: " );
		report.append( m_timedate );
		report.append( " (hex: " );
		report.append( NumUtil.hex2String( m_scts ) );
		report.append( ")" );
		report.append( newline );
		
		if( m_udhi == UDHI_HEADER_PRESENT )
		{
			report.append( "Headers: " );
			report.append( newline );
			for( SMSUserDataHeaderElement e : m_udh )
			{
				report.append( e.toString() );
				report.append( newline );
			}
		}
		
		report.append( "User data: " );
		report.append( m_ud.substring( 0, m_udl ) );
		report.append( newline );
	
		return report.toString();
	}
}
