/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.jaccal.util.NumUtil;
import com.simfill.LogManager;
import com.simfill.Registry;

public class SMSSubmit extends SMSUserData implements SMSReplyPath
{
	// TP-Reject-Duplicates (TP-RD), according to 3GPP TS 23.040
	// Bit 2 of the first byte of a PDU.
	public static final byte RD_ACCEPT            = 0x00;
	public static final byte RD_REJECT            = 0x04;
	
	protected byte m_rd;
	
	// TP-Validity-Period-Format (TP-VPF), according to 3GPP TS 23.040
	// Bits 3 and 4 of the first byte of a PDU.
	// FIXME not supported yet
	public static final byte VPF_NOT_PRESENT      = 0x00;
	public static final byte VPF_RELATIVE         = 0x10;
	public static final byte VPF_ENHANCED         = 0x08;
	public static final byte VPF_ABSOLUTE         = 0x18;
	
	protected byte m_vpf;
	
	// TP-Status-Report-Request (TP-SRR), according to 3GPP TS 23.040
	// Bit 5 of the first byte of a PDU.
	public static final byte SRR_DONT_REPORT      = 0x00;
	public static final byte SRR_DO_REPORT        = 0x20;
	
	protected byte m_srr;
	
	// TP-Reply-Path (TP-RP), according to 3GPP TS 23.040
	// Bit 7 of the first byte of a PDU.
	protected byte m_rp;

	// TP-Message-Reference (TP-MR), according to 3GPP TS 23.040
	protected short m_mr;
	
	// TP-Destination-Address (TP-DA), according to 3GPP TS 23.040
	protected Address m_da;
	
	// TP-Validity-Period (TP-VP), according to 3GPP TS 23.040
	// FIXME not supported yet
	protected byte [] m_vp;
	
	public SMSSubmit( short size ) throws Exception
	{
		super( size );
		
		setMTI( MTI_MS2SC_SMS_SUBMIT );
		setRD ( RD_REJECT );
		setRP ( RP_NOT_SET );
		setVPF( VPF_NOT_PRESENT );
		setSRR( SRR_DONT_REPORT );
		setDA ( new Address() );
	}
	
	public void setRD( byte rd )
	{
		m_rd = rd;
	}
	
	public void setRP( byte rp )
	{
		m_rp = rp;
	}
	
	// FIXME not supported yet
	private void setVPF( byte vpf )
	{
		m_vpf = vpf;
	}
	
	public void setSRR( byte srr )
	{
		m_srr = srr;
	}
	
	public void setMR( short mr )
	{
		m_mr = mr;
	}
	
	public void setDA( Address da )
	{
		m_da = da;
	}
	
	public String toHex()
	{
		byte [] storage = new byte[ getSize() ];

		int idx = 0;

		// Store the status of SMS
		storage[ idx++ ] = m_status;

		// Store the service center number if provided
		String number = m_sca.getNumber();
		if( number.length() == 0 )
		{
			storage[ idx++ ] = 0;
		}
		else
		{
			int size = number.length();
			
			// Check the size of the number
			if( size > 20 )
			{
				LogManager.getInstance().log( "warn", "SMSSubmit: The length of the service center address is limited to 20 digits. The number '" + number + "' will be truncated.", true );
				size = 20;
			}
			
			storage[ idx++ ] = ( byte )( ( size >> 1 ) + ( size & 1 ) + 1 );
			storage[ idx++ ] = m_sca.getTypeOfAddress();
			number += ( char )0xff;
			for( int ch = 0; ch < size; ch += 2 )
			{
				storage[ idx++ ] = ( byte )( ( number.charAt( ch ) - '0' ) + ( ( number.charAt( ch + 1 ) - '0' ) << 4 ) );
			}
		}

		// Store the flags in the first byte of the TDPU
		storage[ idx++ ] = ( byte )( m_mti | m_rd | m_vpf | m_srr | m_udhi | m_rp );
		
		// Store the message reference
		storage[ idx++ ] = ( byte )m_mr;

		// Store the originated address if provided
		number = m_da.getNumber();
		if( number.length() == 0 )
		{
			storage[ idx++ ] = 0;
			storage[ idx++ ] = m_da.getTypeOfAddress();
		}
		else
		{
			int size = number.length();
			
			// Check the size of the number
			if( size > 20 )
			{
				LogManager.getInstance().log( "warn", "SMSSubmit: The length of the originated address is limited to 20 digits. The number '" + number + "' will be truncated.", true );
				size = 20;
			}
			
			storage[ idx++ ] = ( byte )size;
			storage[ idx++ ] = m_da.getTypeOfAddress();
			number += ( char )0xff;
			for( int ch = 0; ch < size; ch += 2 )
			{
				storage[ idx++ ] = ( byte )( ( number.charAt( ch ) - '0' ) + ( ( number.charAt( ch + 1 ) - '0' ) << 4 ) );
			}
		}

		// Store the protocole identifier
		storage[ idx++ ] = m_pid;

		// Store the data coding scheme
		storage[ idx++ ] = m_dcs;

		// Store the validity period
		switch( m_vpf )
		{
			case VPF_RELATIVE:
				
				storage[ idx++ ] = m_vp[ 0 ];
				break;
				
			case VPF_ABSOLUTE:
			case VPF_ENHANCED:
				
				for( int it = 0; it < 7; ++it )
				{
					storage[ idx++ ] = m_vp[ it ];
				}
				break;
				
			case VPF_NOT_PRESENT:
			default:
				
				break;
		}

		// Store the user data header, if necessary
		int udl_idx = idx++;
		
		if( m_udhi == UDHI_HEADER_PRESENT )
		{
			storage[ idx++ ] = ( byte )m_udhl;

			for( int it = 0; it < m_udh.size() && idx < getSize(); ++it )
			{
				SMSUserDataHeaderElement e = m_udh.elementAt( it );
				
				System.arraycopy( e.getData(), 0, storage, idx, e.getSize() );
				idx += e.getSize();
			}
		}
		
		// Store the user data
		try
		{
			int avail;

			switch( m_dcs )
			{
				case DCS_16BITS_UCS2:
					
					// Check the space available
					avail = Math.min( 140 - ( m_udhl + 1 ), getSize() - idx ) >> 1;
					
					// Check the length of the text
					if( m_udl > avail )
					{
						m_udl = avail;
						LogManager.getInstance().log( "warn", "SMSDeliver: The text '" + m_ud + "' is UCS2-encoded. It will be truncated to " + avail + " characters.", true );
					}

					// Store the text as is, as Java stores the text UCS2 encoded already
					for( int ch = 0; ch < m_udl; ++ch )
					{
						char c = m_ud.charAt( ch );

						storage[ idx++ ] = ( byte )( ( c & 0xff00 ) >> 8 );
						storage[ idx++ ] = ( byte )( ( c & 0x00ff ) >> 0 );
					}
					
					// Store the size of the user data, including the header
					storage[ udl_idx ] = ( byte )( idx - udl_idx - 1 );

					break;

				case DCS_7BITS_DEFAULT:

					int offset = 0;
					int boundary = 0;

					// Compute calculation related to 7-bit boundary padding.
					// According to 3GPP TS 23.040, the text has to start on a 7-bit boundary. 
					if( m_udhi == UDHI_HEADER_PRESENT )
					{
						offset = ( m_udhl + 1 ) << 3;								// number of bits used by the headers
						boundary = 7 * ( int )Math.ceil( ( double )offset / 7.0 );	// number of bits to reach the next 7-bit boundary
						avail = Math.min( 160 - boundary / 7, getSize() - idx );	// "avail" is the number of septets available
						offset -= boundary - 7;
					}
					else
					{
						avail = Math.min( 160, getSize() - idx );
					}

					if( m_udl > 0 )
					{
						// Check the length of the text
						byte [] trText = Charset.translate( m_ud );
						m_udl = trText.length;
						if( m_udl > avail )
						{
							m_udl = avail;
							LogManager.getInstance().log( "warn", "SMSDeliver: The text '" + m_ud + "' is 7-bits-encoded and will be truncated.", true );
						}
	
						// If the truncation happens on a 2-bytes special character, also remove the remaining byte
						if( trText[ m_udl - 1 ] == Charset.CHARSET_ESCAPE ) --m_udl;
	
						// Encode the text using the 8 bits to 7 bits compression algorithm
						byte [] padding = { 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f };
						byte [] padText = new byte[ m_udl + padding.length ];
						System.arraycopy( trText,  0, padText, 0, m_udl );
						System.arraycopy( padding, 0, padText, m_udl, padding.length );
	
						int ch = 0;
						int ch1 = ( m_udhi == UDHI_HEADER_PRESENT ) ? 0x7f : padText[ ch++ ]; // Pad the unused bits between the end of the header and the next 7-bit boundary where the text begins
						while( ( ch <= m_udl || offset > 0 ) && idx < getSize() )
						{
							int ch2 = padText[ ch++ ];
							storage[ idx++ ] = ( byte )( ( ch1 >> offset ) + ( ( ch2 & ( 0x000000ff >> ( 7 - offset ) ) ) << ( 7 - offset ) ) );
							if( ++offset < 7 )
							{
								ch1 = ch2;
							}
							else
							{
								offset = 0;
								ch1 = padText[ ch++ ];
							}
						}

						// Store the number of septets of the user data, including the header
						storage[ udl_idx ] = ( byte )Math.ceil( ( boundary + m_udl * 7.0 ) / 7 );
						//storage[ udl_idx ] = ( byte )Math.floor( ( ( idx - udl_idx - 1 ) * 8 ) / 7 );

						// Store the truncated user data
						setUD( Charset.untranslate( trText, 0, m_udl ) );
					}
					else
					{
						// Store the number of septets of the user data, including the header
						storage[ udl_idx ] = ( byte )( boundary / 7 );
					}

					break;
					
				default:
					
					LogManager.getInstance().log( "warn", "SMSDeliver: unsupported data coding scheme (TP-DCS). No user data will be stored in the SMS message.", true );

					// Store the size of the user data, including the header
					storage[ udl_idx ] = ( byte )( idx - udl_idx - 1 );

					break;
			}
		}
		catch( Exception e )
		{
			LogManager.getInstance().log( "excp", e, true );
			return null;
		}

		// Fill the remaining space with padding
		while( idx < getSize() )
		{
			storage[ idx++ ] = -1;
		}

		return NumUtil.hex2String( storage );
	}
	
	public String toString()
	{
		StringBuffer report = new StringBuffer();
		
		String newline = Registry.getInstance().get( "report.newline" );

		report.append( "Status: " );
		switch( m_status )
		{
			case STATUS_FREESPACE:    report.append( "Empty / deleted"    ); break;
			case STATUS_IN_READ:      report.append( "Incoming, read"     ); break;
			case STATUS_IN_NOT_READ:  report.append( "Incoming, not read" ); break;
			case STATUS_OUT_SENT:     report.append( "Outgoing, sent"     ); break;
			case STATUS_OUT_NOT_SENT: report.append( "Outgoing, not sent" ); break;
			default:                  report.append( "INVALID" );
		}
		report.append( newline );
		
		report.append( "Service center address: " );
		report.append( newline );
		if( m_sca.getNumber().length() > 20 )
		{
			Address a = new Address();
			a.setTON( m_sca.getTON() );
			a.setNPI( m_sca.getNPI() );
			a.setNumber( m_sca.getNumber().substring( 0, 20 ) );
			report.append( a );
		}
		else
		{
			report.append( m_sca );
		}
		
		report.append( "Type: " );
		switch( m_mti )
		{
			case MTI_MS2SC_SMS_DELIVER_REPORT: report.append( "SMS-DELIVER-REPORT" ); break;
			case MTI_MS2SC_SMS_SUBMIT:         report.append( "SMS-SUBMIT"         ); break;
			case MTI_MS2SC_SMS_COMMAND:        report.append( "SMS-COMMAND"        ); break;
			case MTI_RESERVED:                 report.append( "Reserved"           ); break;
			default:                           report.append( "INVALID"            );
		}
		report.append( newline );
		
		report.append( "Reject duplicates: " );
		report.append( m_rd == RD_REJECT );
		report.append( newline );
		
		report.append( "Validity period format: " );
		switch( m_vpf )
		{
			case VPF_NOT_PRESENT: report.append( "not present" ); break;
			case VPF_RELATIVE:    report.append( "relative"    ); break;
			case VPF_ENHANCED:    report.append( "enhanced"    ); break;
			case VPF_ABSOLUTE:    report.append( "absolute"    ); break;
			default:              report.append( "INVALID"     );
		}
		report.append( newline );
		
		report.append( "Status report requested: " );
		report.append( m_srr == SRR_DO_REPORT );
		report.append( newline );
		
		report.append( "User data header present: " );
		report.append( m_udhi == UDHI_HEADER_PRESENT );
		report.append( newline );
		
		report.append( "Reply path set: " );
		report.append( m_rp == RP_SET );
		report.append( newline );
		
		report.append( "Message reference: " );
		report.append( m_mr );
		report.append( newline );
		
		report.append( "Destination address: ");
		report.append( newline );
		if( m_da.getNumber().length() > 20 )
		{
			Address a = new Address();
			a.setTON( m_da.getTON() );
			a.setNPI( m_da.getNPI() );
			a.setNumber( m_da.getNumber().substring( 0, 20 ) );
			report.append( a );
		}
		else
		{
			report.append( m_da );
		}
		
		report.append( "Protocol identifier: NOT SUPPORTED" ); // FIXME
		report.append( newline );
		
		report.append( "Data coding scheme: " );
		switch( m_dcs )
		{
			case DCS_7BITS_DEFAULT: report.append( "7 bits (default)" ); break;
			case DCS_8BITS:         report.append( "8 bits"           ); break;
			case DCS_16BITS_UCS2:   report.append( "16 bits (UCS2)"   ); break;
			case DCS_RESERVED:      report.append( "Reserved"         ); break;
			default:                report.append( "INVALID"          );
		}
		report.append( newline );
		
		report.append( "Validity period: NOT SUPPORTED" ); // FIXME
		report.append( newline );
		
		if( m_udhi == UDHI_HEADER_PRESENT )
		{
			report.append( "Headers: " );
			report.append( newline );
			for( SMSUserDataHeaderElement e : m_udh )
			{
				report.append( e.toString() );
				report.append( newline );
			}
		}
		
		report.append( "User data: " );
		report.append( m_ud.substring( 0, m_udl ) );
		report.append( newline );
	
		return report.toString();
	}
}