/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.simfill.LogManager;
import com.simfill.Registry;
import com.jaccal.util.NumUtil;

// Stores the IMSI data to be stored in EF_IMSI
public class SST extends TransparentFile
{
	public SST( int size ) throws Exception
	{
		super( size );
		
		if( getSize() < 2 )
		{
			throw new Exception( "SST: the size of the file EF_SST is not standard." );
		}
		
		m_sst = new byte [ getSize() ];
	}

	public void setService( int number, boolean allocated, boolean activated ) throws Exception
	{
		int byteIndex = ( ( number - 1 ) & 0xfffffffc ) >> 2;
		int bitIndex  = ( ( number - 1 ) & 0x00000003 ) << 1;

		if( byteIndex >= m_sst.length )
		{
			LogManager.getInstance().log( "warn", "SST: according to the size of EF_SST, the maximum service number is " + Integer.toString( getSize() << 2 ) + ". Service n" + Integer.toString( number ) + " is out of bounds.", true );
			throw new Exception( "SST: service out of bounds" );
		}
		
		if( ! allocated && activated )
		{
			LogManager.getInstance().log( "warn", "SST: a service must be allocated to be activated. See service n" + Integer.toString( number ), true );
			throw new Exception( "SST: invalid service status" );
		}
		
		// Reset the two bits of the service
		byte mask = ( byte )( ~( 3 << bitIndex ) );
		m_sst[ byteIndex ] &= mask;
		
		// Set the new two bits
		int lb = allocated ? 1 : 0;
		int hb = activated ? 2 : 0;
		m_sst[ byteIndex ] |= ( byte )( ( lb | hb ) << bitIndex ); 
	}

	public String toHex()
	{
		return NumUtil.hex2String( m_sst );
	}
	
	public String toString()
	{
		String newline = Registry.getInstance().get( "report.newline" );
		StringBuilder storage = new StringBuilder();

		for( int byteIndex = 0; byteIndex < m_sst.length; ++byteIndex )
		{
			for( int bitIndex = 0; bitIndex < 8; bitIndex += 2 )
			{
				storage.append( "Service n" );
				storage.append( Integer.toString( ( byteIndex << 2 ) + ( bitIndex >> 1 ) + 1 ) );
				storage.append( ( m_sst[ byteIndex ] & ( 1 << ( bitIndex + 0 ) ) ) != 0 ? ": allocated" : ": not allocated" );
				storage.append( ( m_sst[ byteIndex ] & ( 1 << ( bitIndex + 1 ) ) ) != 0 ? ", activated" : ", not activated" );
				storage.append( newline );
			}
		}
		
		return storage.toString();
	}

	// Storage for the service table
	private byte [] m_sst;
}
